from . import tools


def set_primary_variable_lower_range_value(address: bytes, value) -> bytes:
    return tools.pack_command(address, command_id=37)


def reset_configuration_changed_flag(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=38)


def perform_master_reset(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=42)


def read_additional_transmitter_status(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=48)


def read_dynamic_variable_assignments(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=50)


def write_number_of_response_preambles(address: bytes, number: int) -> bytes:
    data = number.to_bytes(1, "big")  # check for bytes in manual
    return tools.pack_command(address, command_id=59, data=data)


def toggle_analog_output_mode(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=66)


def trim_analog_output_zero(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=67)


def trim_analog_output_span(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=68)


def select_baud_rate(address: bytes, rate: int) -> bytes:
    data = rate.to_bytes(1, "big")  # check bytes in manual
    return tools.pack_command(address, command_id=123, data=data)



import math
from typing import Union


def calculate_checksum(command: Union[int, bytes]) -> bytes:
    if type(command) == int:
        command = command.to_bytes(64, "big")  # type: ignore
    lrc = 0
    for byte in command:  # type: ignore
        lrc ^= byte
    out = lrc.to_bytes(1, "big")
    return out


def calculate_long_address(manufacturer_id: int, manufacturer_device_type: int, device_id: bytes):
    out = int.from_bytes(device_id, "big")
    out |= manufacturer_device_type << 24
    out |= manufacturer_id << 32
    return out.to_bytes(5, "big")


def pack_command(address, command_id, data=None):
    if type(address) == bytes:
        address = int.from_bytes(address, "big")
    if type(command_id) == int:
        command_id = command_id.to_bytes(1, "big")
    command = b"\xFF\xFF\xFF\xFF\xFF"  # preamble
    command += b"\x82"  # start charachter
    command += (549755813888 | address).to_bytes(5, "big")
    command += command_id
    if data is None:
        command += b"\x00"  # byte count
    else:
        command += len(data).to_bytes(1, "big")  # byte count
        command += data  # data
    command += calculate_checksum(command[5:])
    return command


def pack_ascii(string: Union[str, bytes]) -> bytes:
    if type(string) == str:
        chars = [c.encode() for c in string]  # type: ignore
    else:
        chars = [c for c in string]  # type: ignore
    out = 0
    for i, c in zip(range(8), [ord(c) & 0b0011_1111 for c in chars][::-1]):
        out |= c << (i * 6)
    return out.to_bytes(math.ceil((len(string) * 6) / 8), "big")



import struct
from typing import MutableMapping, Union


def parse(response: bytes) -> MutableMapping[str, Union[int, bytes, str, float]]:
    out: MutableMapping[str, Union[int, bytes, str, float]] = dict()
    out["full_response"] = response
    if response[0] & 0x80:  # long address
        out["address"] = int.from_bytes(response[1:6], "big")
        response = response[6:]
    else:  # short address
        out["address"] = response[1]
        response = response[2:]
    command, bytecount, response_code, device_status = struct.unpack_from(">BBBB", response)
    out["device_status"] = device_status
    out["response_code"] = response_code
    data = response[4 : 4 + bytecount]
    out["command"] = command
    out["command_name"] = f"hart_command_{command}"
    out["bytecount"] = bytecount
    out["data"] = data

    # handle error return
    if bytecount == 2:
        return out

    # universal commands
    if command in [0, 11]:
        out["command_name"] = "read_unique_identifier"
        out["manufacturer_id"] = data[1]
        out["manufacturer_device_type"] = data[2]
        out["number_response_preamble_characters"] = data[3]
        out["universal_command_revision_level"] = data[4]
        out["transmitter_specific_command_revision_level"] = data[5]
        out["software_revision_level"] = data[6]
        out["hardware_revision_level"] = data[7]
        out["device_id"] = int.from_bytes(data[9:12], "big")
    elif command in [1]:
        out["command_name"] = "read_primary_variable"
        units, variable = struct.unpack_from(">Bf", data)
        out["primary_variable_units"] = units
        out["primary_variable"] = variable
    elif command in [2]:
        out["command_name"] = "read_loop_current_and_percent"
        analog_signal, primary_variable = struct.unpack_from(">ff", data)
        out["analog_signal"] = analog_signal
        out["primary_variable"] = primary_variable
    elif command in [3]:
        out["command_name"] = "read_dynamic_variables_and_loop_current"
        (
            analog_signal,
            primary_variable_units,
            primary_variable,
            secondary_variable_units,
            secondary_variable,
        ) = struct.unpack_from(">fBfBf", data)
        out["analog_signal"] = analog_signal
        out["primary_variable_units"] = primary_variable_units
        out["primary_variable"] = primary_variable
        out["secondary_variable_units"] = secondary_variable_units
        out["secondary_variable"] = secondary_variable
    elif command in [6]:
        out["command_name"] = "write_polling_address"
        polling_address = struct.unpack_from(">B", data)[0]
        out["polling_address"] = polling_address
    elif command in [12]:
        out["command_name"] = "read_message"
        out["message"] = data[0:23]
    elif command in [13]:
        out["command_name"] = "read_tag_descriptor_date"
        out["device_tag_name"] = data[0:5]
        out["device_descriptor"] = data[6:17]
        out["date"] = data[18:20]
    elif command in [14]:
        out["command_name"] = "read_primary_variable_information"
        out["serial_no"] = data[0:2]
        sensor_limits_code, upper_limit, lower_limit, min_span = struct.unpack_from(
            ">xxxBfff", data
        )
        out["sensor_limits_code"] = sensor_limits_code
        out["upper_limit"] = upper_limit
        out["lower_limit"] = lower_limit
        out["min_span"] = min_span
    elif command in [15]:
        out["command_name"] = "read_output_information"
        (
            alarm_code,
            transfer_fn_code,
            primary_variable_range_code,
            upper_range_value,
            lower_range_value,
            damping_value,
            write_protect,
            private_label,
        ) = struct.unpack_from(">BBBfffBB", data)
        out["alarm_code"] = alarm_code
        out["transfer_fn_code"] = transfer_fn_code
        out["primary_variable_range_code"] = primary_variable_range_code
        out["upper_range_value"] = upper_range_value
        out["lower_range_value"] = lower_range_value
        out["damping_value"] = damping_value
        out["write_protect"] = write_protect
        out["private_label"] = private_label
    elif command in [16]:
        out["command_name"] = "read_final_assembly_number"
        out["final_assembly_no"] = int.from_bytes(data[0:2], "big")
    elif command in [17]:
        out["command_name"] = "write_message"
        out["message"] = data[0:23]
    elif command in [18]:
        out["command_name"] = "write_tag_descriptor_date"
        out["device_tag_name"] = data[0:5]
        out["device_descriptor"] = data[6:17]
        out["date"] = data[18:20]
    elif command in [19]:
        out["command_name"] = "write_final_assembly_number"
        out["final_assembly_no"] = int.from_bytes(data[0:2], "big")

    # COMMON COMMANDS

    # elif command in [37]:
    #     out["command_name"] = "set_primary_variable_lower_range_value"
    #     out[""] =
    # request data bytes = NONE, response data bytes = NONE

    # elif command in [38]:
    #     out["command_name"] = "reset_configuration_changed_flag"
    #     out[""] =
    # request data bytes = NONE, response data bytes = NONE

    # elif command in [42]:
    #     out["command_name"] = "perform_master_reset"
    #     out[""] =
    # request data bytes = NONE, response data bytes = NONE

    # elif command in [48]:
    #     out["command_name"] = "read_additional_transmitter_status"
    #     out[""] =
    # request data bytes = NONE, response data bytes = NONE

    elif command in [50]:
        out["command_name"] = "read_dynamic_variable_assignments"
        (
            primary_transmitter_variable,
            secondary_transmitter_variable,
            tertiary_transmitter_variable,
            quaternary_transmitter_variable,
        ) = struct.unpack_from(">BBBB", data)
        out["primary_transmitter_variable"] = primary_transmitter_variable
        out["secondary_transmitter_variable"] = secondary_transmitter_variable
        out["tertiary_transmitter_variable"] = tertiary_transmitter_variable  # NOT USED
        out["quaternary_transmitter_variable"] = quaternary_transmitter_variable  # NOT USED
    elif command in [59]:
        out["command_name"] = "write_number_of_response_preambles"
        n_response_preambles = struct.unpack_from(">B", data)[0]
        out["n_response_preambles"] = n_response_preambles
    elif command in [66]:
        out["command_name"] = "toggle_analog_output_mode"
        (
            analog_output_selection,
            analog_output_units_code,
            fixed_analog_output,
        ) = struct.unpack_from(">BBf", data)
        out["analog_output_selection"] = analog_output_selection
        out["analog_output_units_code"] = analog_output_units_code
        out["fixed_analog_output"] = fixed_analog_output
    elif command in [67]:
        out["command_name"] = "trim_analog_output_zero"
        analog_output_code, analog_output_units_code, measured_analog_output = struct.unpack_from(
            ">BBf", data
        )
        out["analog_output_code"] = analog_output_code
        out["analog_output_units_code"] = analog_output_units_code
        out["measured_analog_output"] = measured_analog_output
    elif command in [68]:
        out["command_name"] = "trim_analog_output_span"
        analog_output_code, analog_output_units_code, measured_analog_output = struct.unpack_from(
            ">BBf", data
        )
        out["analog_output_code"] = analog_output_code
        out["analog_output_units_code"] = analog_output_units_code
        out["measured_analog_output"] = measured_analog_output
    elif command in [123]:
        out["command_name"] = "select_baud_rate"
        out["baud_rate"] = int.from_bytes(data, "big")

    return out



from . import tools


def read_unique_identifier(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=0)


def read_primary_variable(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=1)


def read_loop_current_and_percent(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=2)


def read_dynamic_variables_and_loop_current(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=3)


def write_polling_address(address: bytes, new_polling_address: int) -> bytes:
    assert 0 <= new_polling_address <= 15
    return tools.pack_command(address, command_id=6, data=new_polling_address.to_bytes(1, "big"))


def read_unique_identifier_associated_with_tag(tag: bytes, *, address: int = 0) -> bytes:
    return tools.pack_command(address, command_id=11, data=tag)


def read_message(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=12)


def read_tag_descriptor_date(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=13)


def read_primary_variable_information(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=14)


def read_output_information(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=15)


def read_final_assembly_number(address: bytes) -> bytes:
    return tools.pack_command(address, command_id=16)


def write_message(address: bytes, message: str) -> bytes:
    message = message.ljust(32)
    return tools.pack_command(address, command_id=17, data=tools.pack_ascii(message))


def write_tag_descriptor_date(address: bytes, tag: str, descriptor: str, date: tuple):
    data = b""
    assert len(tag) <= 8
    data += tools.pack_ascii(tag.ljust(8))
    assert len(descriptor) <= 16
    data += tools.pack_ascii(descriptor.ljust(16))
    day, month, year = date
    data += day.to_bytes(1, "big")
    data += month.to_bytes(1, "big")
    data += year.to_bytes(1, "big")
    return tools.pack_command(address, command_id=18, data=data)


def write_final_assembly_number(address: bytes, number: int):
    data = number.to_bytes(3, "big")
    return tools.pack_command(address, command_id=19, data=data)



__all__ = ["Unpacker"]

import asyncio
from collections import namedtuple
import io
import struct
import warnings

from ._parsing import parse
from . import tools


class Unpacker:
    """
    Create an Unpacker to decode a byte stream into HART protocol messages.

    The ``file_like`` parameter should be an object which data can be sourced from.
    It should support the ``read()`` method.

    The ``on_error`` parameter selects the action to take if invalid data is detected.
    If set to ``"continue"`` (the default), bytes will be discarded if the byte sequence
    does not appear to be a valid message.
    If set to ``"warn"``, the behaviour is identical, but a warning message will be emitted.
    To instead immediately abort the stream decoding and raise a ``RuntimeError``, set to
    ``"raise"``.

    :param file_like: A file-like object which data can be `read()` from.
    :param on_error: Action to take if invalid data is detected.
    """

    def __init__(self, file_like=None, on_error="continue"):
        if file_like is None:
            self._file = io.BytesIO()
        else:
            self._file = file_like
        self.buf = b""
        self.on_error = on_error

    def __iter__(self):
        return self

    def _decoding_error(self, message="Error decoding message from buffer."):
        """
        Take appropriate action if parsing of data stream fails.

        :param message: Warning or error message string.
        """
        if self.on_error == "raise":
            raise RuntimeError(message)
        if self.on_error == "warn":
            warnings.warn(message)

    def _read_one_byte_if_possible(self):
        if self._file.in_waiting > 0:
            return self._file.read(1)
        else:
            raise StopIteration

    def __next__(self):
        # must work with at least two bytes to start with
        while len(self.buf) < 3:
            self.buf += self._read_one_byte_if_possible()
        # keep reading until we find a minimum preamble
        while self.buf[:3] not in [b"\xFF\xFF\x06", b"\xFF\xFF\x86"]:
            self.buf += self._read_one_byte_if_possible()
            self.buf = self.buf[1:]
            self._decoding_error("Head of buffer not recognized as valid preamble")
        # now the head of our buffer is the start charachter plus two preamble
        # we will read all the way through status
        if self.buf[2] & 0x80:
            l = 12
        else:
            l = 8
        while len(self.buf) < l:
            self.buf += self._read_one_byte_if_possible()
        # now we can use the bytecount to read through the data and checksum
        bytecount = self.buf[l - 3]
        response_length = l + bytecount - 1
        while len(self.buf) < response_length:
            self.buf += self._read_one_byte_if_possible()
        # checksum
        checksum = int.from_bytes(
            tools.calculate_checksum(self.buf[2 : response_length - 1]), "big"
        )
        if checksum != self.buf[response_length - 1]:
            self._decoding_error("Invalid checksum.")
            raise StopIteration
        # parse
        response = self.buf[2:response_length]
        dict_ = parse(response)
        # clear buffer
        if len(self.buf) == response_length:
            self.buf = b""
        else:
            self.buf = self.buf[response_length + 3 :]
        # return
        return namedtuple(dict_["command_name"], dict_.keys())(**dict_)

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                return next(self)
            except StopIteration:
                await asyncio.sleep(0.001)

    def feed(self, data: bytes):
        """
        Add byte data to the input stream.

        The input stream must support random access, if it does not, must be fed externally
        (e.g. serial port data).

        :param data: Byte array containing data to add.
        """
        pos = self._file.tell()
        self._file.seek(0, 2)
        self._file.write(data)
        self._file.seek(pos)



"""A sans-io python implementation of the Highway Addressable Remote Transducer Protocol."""

from .__version__ import *

from ._unpacker import *
from . import common
from . import universal
from . import tools
